import pytest
from baby_steps import given, then, when
from pytest import raises
from vedro.core import Dispatcher
from vedro.core import MonotonicScenarioScheduler as Scheduler
from vedro.core import Report
from vedro.events import CleanupEvent, StartupEvent

from ._utils import dispatcher, fire_arg_parsed_event, fire_startup_event, make_vscenario, tagger

__all__ = ("dispatcher", "tagger")  # fixtures


@pytest.mark.usefixtures(tagger.__name__)
async def test_no_tags(*, dispatcher: Dispatcher):
    with given:
        await fire_arg_parsed_event(dispatcher, tags=None)

        scenarios = [make_vscenario(), make_vscenario()]
        scheduler = Scheduler(scenarios)
        startup_event = StartupEvent(scheduler)

    with when:
        await dispatcher.fire(startup_event)

    with then:
        assert list(scheduler.scheduled) == scenarios


@pytest.mark.usefixtures(tagger.__name__)
async def test_nonexisting_tag(*, dispatcher: Dispatcher):
    with given:
        await fire_arg_parsed_event(dispatcher, tags="SMOKE")

        scenarios = [make_vscenario(), make_vscenario()]
        scheduler = Scheduler(scenarios)
        startup_event = StartupEvent(scheduler)

    with when:
        await dispatcher.fire(startup_event)

    with then:
        assert list(scheduler.scheduled) == []


@pytest.mark.usefixtures(tagger.__name__)
async def test_tag(*, dispatcher: Dispatcher):
    with given:
        await fire_arg_parsed_event(dispatcher, tags="SMOKE")

        scenarios = [
            make_vscenario(tags=["SMOKE"]),
            make_vscenario(),
            make_vscenario(tags=["SMOKE", "P0"])
        ]
        scheduler = Scheduler(scenarios)
        startup_event = StartupEvent(scheduler)

    with when:
        await dispatcher.fire(startup_event)

    with then:
        assert list(scheduler.scheduled) == [scenarios[0], scenarios[2]]


@pytest.mark.usefixtures(tagger.__name__)
async def test_tag_not_operator(*, dispatcher: Dispatcher):
    with given:
        await fire_arg_parsed_event(dispatcher, tags="not SMOKE")

        scenarios = [
            make_vscenario(tags=["SMOKE"]),
            make_vscenario(),
            make_vscenario(tags=["SMOKE", "P0"])
        ]
        scheduler = Scheduler(scenarios)
        startup_event = StartupEvent(scheduler)

    with when:
        await dispatcher.fire(startup_event)

    with then:
        assert list(scheduler.scheduled) == [scenarios[1]]


@pytest.mark.usefixtures(tagger.__name__)
async def test_tag_and_operator(*, dispatcher: Dispatcher):
    with given:
        await fire_arg_parsed_event(dispatcher, tags="SMOKE and P0")

        scenarios = [
            make_vscenario(tags=["SMOKE"]),
            make_vscenario(),
            make_vscenario(tags=["SMOKE", "P0"]),
            make_vscenario(tags=["P0"])
        ]
        scheduler = Scheduler(scenarios)
        startup_event = StartupEvent(scheduler)

    with when:
        await dispatcher.fire(startup_event)

    with then:
        assert list(scheduler.scheduled) == [scenarios[2]]


@pytest.mark.usefixtures(tagger.__name__)
async def test_tag_or_operator(*, dispatcher: Dispatcher):
    with given:
        await fire_arg_parsed_event(dispatcher, tags="SMOKE or P0")

        scenarios = [
            make_vscenario(tags=["SMOKE"]),
            make_vscenario(),
            make_vscenario(tags=["SMOKE", "P0"])
        ]
        scheduler = Scheduler(scenarios)
        startup_event = StartupEvent(scheduler)

    with when:
        await dispatcher.fire(startup_event)

    with then:
        assert list(scheduler.scheduled) == [scenarios[0], scenarios[2]]


@pytest.mark.usefixtures(tagger.__name__)
async def test_tags_expr(*, dispatcher: Dispatcher):
    with given:
        await fire_arg_parsed_event(dispatcher, tags="(not SMOKE) and (not P0)")

        scenarios = [
            make_vscenario(tags=["SMOKE"]),
            make_vscenario(),
            make_vscenario(tags=["SMOKE", "P0"])
        ]
        scheduler = Scheduler(scenarios)
        startup_event = StartupEvent(scheduler)

    with when:
        await dispatcher.fire(startup_event)

    with then:
        assert list(scheduler.scheduled) == [scenarios[1]]


@pytest.mark.usefixtures(tagger.__name__)
async def test_tags_skipped(*, dispatcher: Dispatcher):
    with given:
        await fire_arg_parsed_event(dispatcher, tags="SMOKE")

        scenarios = [
            make_vscenario(tags=["SMOKE"], is_skipped=True),
            make_vscenario(is_skipped=True),
            make_vscenario(),
        ]
        scheduler = Scheduler(scenarios)
        startup_event = StartupEvent(scheduler)

    with when:
        await dispatcher.fire(startup_event)

    with then:
        assert list(scheduler.scheduled) == [scenarios[0]]


@pytest.mark.usefixtures(tagger.__name__)
async def test_tags_type_validation(*, dispatcher: Dispatcher):
    with given:
        await fire_arg_parsed_event(dispatcher, tags="SMOKE")

        scenario = make_vscenario(tags={"SMOKE": "SMOKE"})  # type: ignore
        scheduler = Scheduler([scenario])

        startup_event = StartupEvent(scheduler)

    with when, raises(BaseException) as exc:
        await dispatcher.fire(startup_event)

    with then:
        assert exc.type is TypeError
        assert str(exc.value) == (
            f"Scenario '{scenario.rel_path}' tags must be a list, tuple or set, got <class 'dict'>"
        )


@pytest.mark.usefixtures(tagger.__name__)
async def test_tags_value_validation(*, dispatcher: Dispatcher):
    with given:
        await fire_arg_parsed_event(dispatcher, tags="SMOKE")

        scenario = make_vscenario(tags=["-SMOKE"])
        scheduler = Scheduler([scenario])

        startup_event = StartupEvent(scheduler)

    with when, raises(BaseException) as exc:
        await dispatcher.fire(startup_event)

    with then:
        assert exc.type is ValueError
        assert str(exc.value) == (
            f"Scenario '{scenario.rel_path}' tag '-SMOKE' is not a valid identifier"
        )


@pytest.mark.usefixtures(tagger.__name__)
async def test_tags_show_parsed_disabled(*, dispatcher: Dispatcher):
    with given:
        await fire_arg_parsed_event(dispatcher, tags="API or CLI")
        await fire_startup_event(dispatcher, Scheduler([]))

        report = Report()
        cleanup_event = CleanupEvent(report)

    with when:
        await dispatcher.fire(cleanup_event)

    with then:
        assert report.summary == []


async def test_tags_show_parsed_enabled(*, tagger, dispatcher: Dispatcher):
    with given:
        tagger._show_parsed = True
        await fire_arg_parsed_event(dispatcher, tags="API or CLI")
        await fire_startup_event(dispatcher, Scheduler([]))

        report = Report()
        cleanup_event = CleanupEvent(report)

    with when:
        await dispatcher.fire(cleanup_event)

    with then:
        assert report.summary == [
            '--tags "API or CLI" -> OR(Tag(API), Tag(CLI))'
        ]



from typing import Any, Callable, Type, Union

from vedro.core import Dispatcher, Plugin, PluginConfig, VirtualScenario
from vedro.events import ArgParsedEvent, ArgParseEvent, CleanupEvent, StartupEvent

from ._tag_matcher import AdvancedTagMatcher, TagMatcher

__all__ = ("VedroAdvancedTags", "VedroAdvancedTagsPlugin",)


class VedroAdvancedTagsPlugin(Plugin):
    def __init__(self, config: Type["VedroAdvancedTags"], *,
                 tag_matcher_factory: Callable[[str], TagMatcher] = AdvancedTagMatcher) -> None:
        super().__init__(config)
        self._show_parsed = config.show_parsed
        self._matcher_factory = tag_matcher_factory
        self._matcher: Union[TagMatcher, None] = None
        self._tags: Union[str, None] = None

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ArgParseEvent, self.on_arg_parse) \
            .listen(ArgParsedEvent, self.on_arg_parsed) \
            .listen(StartupEvent, self.on_startup) \
            .listen(CleanupEvent, self.on_cleanup)

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        event.arg_parser.add_argument("-t", "--tags", help="Set tags")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._tags = event.args.tags

    def _validate_tags(self, scenario: VirtualScenario, tags: Any) -> bool:
        if not isinstance(tags, (list, tuple, set)):
            raise TypeError(f"Scenario '{scenario.rel_path}' tags must be a list, tuple or set, "
                            f"got {type(tags)}")

        for tag in tags:
            if not tag.isidentifier():
                raise ValueError(
                    f"Scenario '{scenario.rel_path}' tag '{tag}' is not a valid identifier")

        return True

    async def on_startup(self, event: StartupEvent) -> None:
        if self._tags is None:
            return

        self._matcher = self._matcher_factory(self._tags)
        async for scenario in event.scheduler:
            tags = getattr(scenario._orig_scenario, "tags", ())
            self._validate_tags(scenario, tags)
            if not self._matcher.match(list(tags)):
                event.scheduler.ignore(scenario)

    def on_cleanup(self, event: CleanupEvent) -> None:
        if self._tags and self._matcher and self._show_parsed:
            event.report.add_summary(
                f'--tags "{self._tags}" -> {self._matcher._grammar!r}')  # type: ignore


class VedroAdvancedTags(PluginConfig):
    plugin = VedroAdvancedTagsPlugin

    # Show parsed tags in summary
    show_parsed: bool = False



from abc import ABC, abstractmethod

from ._grammar import TagsType, parse

__all__ = ("AdvancedTagMatcher", "TagMatcher",)


class TagMatcher(ABC):
    def __init__(self, expr: str) -> None:
        self._expr = expr

    @abstractmethod
    def match(self, tags: TagsType) -> bool:
        pass


class AdvancedTagMatcher(TagMatcher):
    def __init__(self, expr: str) -> None:
        super().__init__(expr)
        self._grammar = parse(expr)

    def match(self, tags: TagsType) -> bool:
        return self._grammar(tags)



from baby_steps import given, then, when
from pyparsing import ParseResults

from vedro_advanced_tags._grammar import AndTag, NotTag, OrTag, TagExpr, TagOperand

from ._utils import make_tag


def test_tag_operand():
    with given:
        tokens = ParseResults(["API"])

    with when:
        tag_operand = TagOperand("...", 0, tokens)

    with then:
        assert isinstance(tag_operand, TagExpr)
        assert repr(tag_operand) == "Tag(API)"


def test_and_operator():
    with given:
        tokens = ParseResults([
            [make_tag("API"), "and", make_tag("P0")]
        ])

    with when:
        and_operator = AndTag("...", 0, tokens)

    with then:
        assert isinstance(and_operator, TagExpr)
        assert repr(and_operator) == "AND(Tag(API), Tag(P0))"


def test_or_operator():
    with given:
        tokens = ParseResults([
            [make_tag("API"), "or", make_tag("P0")]
        ])

    with when:
        or_operator = OrTag("...", 0, tokens)

    with then:
        assert isinstance(or_operator, TagExpr)
        assert repr(or_operator) == "OR(Tag(API), Tag(P0))"


def test_not_operator():
    with given:
        tokens = ParseResults([
            ["not", make_tag("API")]
        ])

    with when:
        not_operator = NotTag("...", 0, tokens)

    with then:
        assert isinstance(not_operator, TagExpr)
        assert repr(not_operator) == "NOT(Tag(API))"



from abc import ABC, abstractmethod
from typing import Callable, List, cast

from pyparsing import Literal, ParseResults, Word, alphanums, infixNotation, opAssoc

__all__ = ("parse", "TagsType",
           "AndTag", "OrTag", "NotTag",
           "TagOperand", "TagOperator",
           "TagExpr", "TagsType",)


TagsType = List[str]


class TagExpr(ABC):
    def __init__(self, orig: str, location: int, tokens: ParseResults) -> None:
        pass

    @abstractmethod
    def __call__(self, tags: TagsType) -> bool:
        pass

    @abstractmethod
    def __repr__(self) -> str:
        pass


class TagOperator(TagExpr):
    pass


class TagOperand(TagExpr):
    def __init__(self, orig: str, location: int, tokens: ParseResults) -> None:
        super().__init__(orig, location, tokens)
        self._value = tokens[0]

    def __repr__(self) -> str:
        return f"Tag({self._value})"

    def __call__(self, tags: TagsType) -> bool:
        return self._value in tags


class AndTag(TagOperator):
    def __init__(self, orig: str, location: int, tokens: ParseResults) -> None:
        super().__init__(orig, location, tokens)
        self._left = tokens[0][0]
        self._right = tokens[0][-1]

    def __repr__(self) -> str:
        return f"AND({self._left}, {self._right})"

    def __call__(self, tags: TagsType) -> bool:
        return cast(bool, self._left(tags) and self._right(tags))


class OrTag(TagOperator):
    def __init__(self, orig: str, location: int, tokens: ParseResults) -> None:
        super().__init__(orig, location, tokens)
        self._left = tokens[0][0]
        self._right = tokens[0][-1]

    def __repr__(self) -> str:
        return f"OR({self._left}, {self._right})"

    def __call__(self, tags: TagsType) -> bool:
        return cast(bool, self._left(tags) or self._right(tags))


class NotTag(TagOperator):
    def __init__(self, orig: str, location: int, tokens: ParseResults) -> None:
        super().__init__(orig, location, tokens)
        self._value = tokens[0][-1]

    def __repr__(self) -> str:
        return f"NOT({self._value})"

    def __call__(self, tags: TagsType) -> bool:
        return not self._value(tags)


_operand = Word(alphanums).setParseAction(TagOperand)
_grammar = infixNotation(_operand, [
    (Literal("not"), 1, opAssoc.RIGHT, NotTag),
    (Literal("and"), 2, opAssoc.LEFT, AndTag),
    (Literal("or"), 2, opAssoc.LEFT, OrTag),
])


def parse(expr: str) -> Callable[[TagsType], bool]:
    return cast(Callable[[TagsType], bool], _grammar.parseString(expr)[0])
